@@ -1,10 +1,11 @@
 package BenchmarkAnything::Storage::Backend::SQL;
-# git description: v0.019-1-g809607b
+# git description: v0.020-21-g75efc23
 
 our $AUTHORITY = 'cpan:TAPPER';
 # ABSTRACT: Autonomous SQL backend to store benchmarks
-$BenchmarkAnything::Storage::Backend::SQL::VERSION = '0.020';
+$BenchmarkAnything::Storage::Backend::SQL::VERSION = '0.021';
 use 5.008;
+use utf8;
 use strict;
 use warnings;
 
@@ -25,6 +26,17 @@ my $hr_default_config = {
     },
 };
 
+my $hr_column_ba_mapping = {
+    bench_value_id => 'VALUE_ID',
+    bench          => 'NAME',
+    bench_value    => 'VALUE',
+    bench_unit     => 'UNIT',
+    created_at     => 'CREATED',
+};
+
+sub json_true  { bless( do{\(my $o = 1)}, 'JSON::PP::Boolean' ) }
+sub json_false { bless( do{\(my $o = 0)}, 'JSON::PP::Boolean' ) }
+
 my $fn_add_subsumed_point = sub {
 
     my ( $or_self, $hr_atts ) = @_;
@@ -152,16 +164,22 @@ sub new {
         });
     }
 
+    $or_self->{searchengine} = $hr_atts->{searchengine} if $hr_atts->{searchengine};
+    $or_self->{debug}        = $hr_atts->{debug} || 0;
+
     return $or_self;
 
 }
 
+
 sub add_single_benchmark {
 
     my ( $or_self, $hr_benchmark, $hr_options ) = @_;
 
     my $hr_config = $or_self->{config};
 
+    my $VALUE_ID; # same spelling as reserved key in BenchmarkAnything schema
+
     # benchmark
     my $i_benchmark_id;
     if ( $hr_benchmark->{NAME} ) {
@@ -240,6 +258,7 @@ sub add_single_benchmark {
                 $hr_config->{tables}{benchmark_value_table},
                 'bench_value_id',
             );
+            $VALUE_ID = $i_benchmark_value_id;
 
             ADDITIONAL: for my $s_key ( keys %{$hr_point} ) {
 
@@ -341,6 +360,26 @@ sub add_single_benchmark {
         return 0;
     }
 
+    if ( $or_self->{searchengine}{elasticsearch}{index_single_added_values_immediately} )
+    {
+        require BenchmarkAnything::Storage::Search::Elasticsearch;
+        my ($or_es, $s_index, $s_type) = BenchmarkAnything::Storage::Search::Elasticsearch::get_elasticsearch_client
+         (
+          {searchengine => $or_self->{searchengine}, ownjson => 1}
+         );
+
+        # Sic, we re-read from DB to get the very same data we
+        # *really got* stored, not just what we wish it should
+        # have stored. That gives us translations like
+        # num->string, CREATED date, etc., etc.
+
+        my $hr_bmk = $or_self->get_single_benchmark_point($VALUE_ID);
+        my $ret = $or_es->index(index => $s_index,
+                                type  => $s_type,
+                                id    => $VALUE_ID,
+                                body  => $hr_bmk);
+    }
+
     return 1;
 
 }
@@ -414,6 +453,14 @@ sub gc {
     my ( $or_self, $hr_options ) = @_;
 
     $or_self->{query}->delete_processed_raw_bench_bundles;
+    if ($or_self->{searchengine}{elasticsearch}{enable_query}) {
+        require BenchmarkAnything::Storage::Search::Elasticsearch;
+        my ($or_es, $s_index, $s_type) = BenchmarkAnything::Storage::Search::Elasticsearch::get_elasticsearch_client
+         (
+          {searchengine => $or_self->{searchengine}}
+         );
+        $or_es->indices->clear_cache(index => $s_index);
+    }
 }
 
 sub add_multi_benchmark {
@@ -536,12 +583,35 @@ sub get_stats {
     my $i_count_datapoints    = $or_self->{query}->select_count_datapoints->fetch->[0];
     my $i_count_metrics       = $or_self->{query}->select_count_metrics->fetch->[0];
     my $i_count_keys          = $or_self->{query}->select_count_keys->fetch->[0];
+    my %h_searchengine_stats  = ();
+
+    # Not strictly *stats* but useful information.
+    if ( $or_self->{searchengine}{elasticsearch}{index} )
+    {
+        require BenchmarkAnything::Storage::Search::Elasticsearch;
+        my ($or_es, $s_index, $s_type) = BenchmarkAnything::Storage::Search::Elasticsearch::get_elasticsearch_client
+         (
+          {searchengine => $or_self->{searchengine}}
+         );
+        %h_searchengine_stats =
+            (
+             elasticsearch =>
+             {
+                 index          => $or_self->{searchengine}{elasticsearch}{index} || 'UNKNOWN',
+                 type           => $or_self->{searchengine}{elasticsearch}{type}  || 'UNKNOWN',
+                 enable_query   => $or_self->{searchengine}{elasticsearch}{enable_query} || 0,
+                 cluster_health => $or_es->cluster->health,
+                 index_single_added_values_immediately => $or_self->{searchengine}{elasticsearch}{index_single_added_values_immediately} || 0,
+             },
+            );
+    }
 
     return {
         count_datapointkeys => 0+$i_count_datapointkeys,
         count_datapoints    => 0+$i_count_datapoints,
         count_metrics       => 0+$i_count_metrics,
         count_keys          => 0+$i_count_keys,
+        %h_searchengine_stats,
     };
 }
 
@@ -577,7 +647,7 @@ sub get_single_benchmark_point {
     $hr_result->{NAME}  = $hr_essentials->{bench};
     $hr_result->{VALUE} = $hr_essentials->{bench_value};
     $hr_result->{VALUE_ID} = $hr_essentials->{bench_value_id};
-    $hr_result->{CREATED_AT} = $hr_essentials->{created_at};
+    $hr_result->{CREATED} = $hr_essentials->{created_at};
     $hr_result->{UNIT}  = $hr_essentials->{bench_unit} if $hr_essentials->{bench_unit};
 
     # cache!
@@ -588,10 +658,62 @@ sub get_single_benchmark_point {
     return $hr_result;
 }
 
+sub get_full_benchmark_points {
+
+    my ( $or_self, $i_bench_value_id, $i_count ) = @_;
+
+    return [] unless $i_bench_value_id;
+
+    $i_count ||= 1;
+
+    # cache?
+    my $s_key;
+    if ( $or_self->{cache} ) {
+        require JSON::XS;
+        $s_key = JSON::XS::encode_json({bench_value_id => $i_bench_value_id});
+        if ( my $hr_search_data = $or_self->{cache}->get("get_full_benchmark_points||$s_key") ) {
+            return $hr_search_data;
+        }
+    }
+
+    # fetch essentials, like NAME, VALUE, UNIT
+    my $ar_essentials = $or_self->{query}
+        ->select_multiple_benchmark_points_essentials($i_bench_value_id, $i_count)
+        ->fetchall_arrayref({});
+    # additional key/value pairs
+    my $ar_additional_values = $or_self->{query}
+        ->select_multiple_benchmark_points_additionals($i_bench_value_id, $i_count)
+        ->fetchall_arrayref({});
+
+    # map columns into BenchmarkAnything schema
+    my $hr_bmk;
+    foreach my $k (keys %$hr_column_ba_mapping)
+    {
+        my $K = $hr_column_ba_mapping->{$k};
+        foreach my $e (@$ar_essentials) {
+            $hr_bmk->{$e->{bench_value_id}}{$K} = $e->{$k} if $k ne 'bench_unit' or defined $e->{$k};
+        }
+    }
+    foreach (@$ar_additional_values) {
+        $hr_bmk->{$_->{bench_value_id}}{$_->{bench_additional_type}} = $_->{bench_additional_value};
+    }
+    # sorted (by VALUE_ID) array of BenchmarkAnything entries
+    my @a_bmk = map { $hr_bmk->{$_} } sort keys %$hr_bmk;
+
+    # cache!
+    if ( $or_self->{cache} ) {
+        $or_self->{cache}->set( "get_full_benchmark_points||$s_key" => \@a_bmk );
+    }
+
+    return \@a_bmk;
+}
+
 sub search_array {
 
     my ( $or_self, $hr_search ) = @_;
 
+    my $debug = $or_self->{debug} || $or_self->{searchengine}{elasticsearch}{debug};
+
     my $s_key;
     if ( $or_self->{cache} ) {
         require JSON::XS;
@@ -601,6 +723,89 @@ sub search_array {
         }
     }
 
+    if ( $debug )
+    {
+        require JSON::XS;
+        require Data::Dumper;
+        print STDERR ',-------------------'."\n";
+        print STDERR "benchmarkanything query:\n";
+        print STDERR "echo '\n";
+        print STDERR JSON::XS->new->pretty->encode($hr_search);
+        print STDERR "' | benchmarkanything-storage search\n";
+        print STDERR '`-------------------'."\n";
+    }
+
+    if ( $or_self->{searchengine}{elasticsearch}{enable_query} )
+    {
+        # If anything goes wrong with Elasticsearch we just continue
+        # below with relational backend query.
+
+        require BenchmarkAnything::Storage::Search::Elasticsearch;
+        my $hr_es_query = BenchmarkAnything::Storage::Search::Elasticsearch::get_elasticsearch_query($hr_search);
+
+        if ($debug)
+        {
+            require JSON::XS;
+            require Data::Dumper;
+            print STDERR ',-------------------'."\n";
+            print STDERR "elasticsearch query:\n";
+            print STDERR "curl -s -XGET 'http://localhost:9200/tapper/benchmarkanything/_search?pretty' -d '\n";
+            print STDERR JSON::XS->new->pretty->encode($hr_es_query);
+            print STDERR "'\n";
+            print STDERR '`-------------------'."\n";
+        }
+
+        # If we could transform the query then we run it against Elasticsearch and return its result.
+        if (defined $hr_es_query)
+        {
+            # ===== client =====
+
+            require BenchmarkAnything::Storage::Search::Elasticsearch;
+            my ($or_es, $s_index, $s_type) = BenchmarkAnything::Storage::Search::Elasticsearch::get_elasticsearch_client
+             (
+              {searchengine => $or_self->{searchengine}, ownjson => 1}
+             );
+
+            # ===== prepare =====
+
+            # If sort fields are of type 'text' then those fields needs to
+            # get their properties being declared as "fielddata":true.
+            my $field_mapping = {};
+            my @sort_fields = map {keys %$_} @{$hr_es_query->{sort}||[]};
+            if (@sort_fields) {
+                $field_mapping = $or_es->indices->get_mapping->{$s_index}{mappings}{$s_type}{properties};
+            }
+            foreach my $sort_field (@sort_fields)
+            {
+                if ($field_mapping->{$sort_field}{type} and $field_mapping->{$sort_field}{type} eq 'text')
+                {
+                    $or_es->indices->put_mapping
+                     (
+                      index => $s_index,
+                      type => $s_type,
+                      body => { $s_type => { properties => { $sort_field => { type => 'text',
+                                                                              fielddata => json_true,
+                                                                            }}}}
+                     );
+                }
+            }
+
+            # ===== search =====
+            my $hr_es_answer = $or_es->search(index => $s_index, type => $s_type, body => $hr_es_query);
+
+            if (
+                !$hr_es_answer->{timed_out} and
+                !$hr_es_answer->{_shards}{failed}
+               )
+            {
+                my @ar_es_result = map { $_->{_source} } @{$hr_es_answer->{hits}{hits} || []};
+                return \@ar_es_result;
+            }
+        }
+
+        # Else no-op, continue with relational backend query.
+    }
+
     my $ar_result = $or_self
         ->search( $hr_search )
         ->fetchall_arrayref({})
@@ -798,6 +1003,109 @@ sub benchmark_operators {
 
 }
 
+sub init_search_engine
+{
+    my ( $or_self, $b_force) = @_;
+
+    my $debug = $or_self->{debug} || $or_self->{searchengine}{elasticsearch}{debug};
+
+    if ( $or_self->{searchengine}{elasticsearch}{index} )
+    {
+        require BenchmarkAnything::Storage::Search::Elasticsearch;
+        my ($or_es, $s_index, $s_type) = BenchmarkAnything::Storage::Search::Elasticsearch::get_elasticsearch_client
+         (
+          {searchengine => $or_self->{searchengine}}
+         );
+
+        # exists?
+        if ($or_es->indices->exists(index => $s_index) and not $b_force)
+        {
+            print STDERR "init_search_engine: index '$s_index' already exists, use force to delete and recreate.\n";
+            return;
+        }
+
+        # delete
+        if ($or_es->indices->exists(index => $s_index))
+        {
+            my $response = $or_es->indices->delete(index => $s_index);
+        }
+
+        # mappings
+        my $mappings =
+        {
+          $s_type => { properties => {
+                                     NAME       => { type         => 'keyword',
+                                                     store        => json_true,
+                                                   },
+                                     VALUE      => { type         => 'text',
+                                                   },
+                                     VALUE_ID   => { type         => 'long',
+                                                   },
+                                     UNIT       => { type         => 'text',
+                                                   },
+                                     CREATED    => { type         => 'date',
+                                                     format       => 'yyyy-MM-dd||yyyy-MM-dd HH:mm:ss',
+                                                   },
+                                    }
+                    },
+        };
+        require JSON::XS;
+        require Hash::Merge;
+        require Data::Dumper;
+        local $Types::Serialiser::true;
+        my $merge = Hash::Merge->new( 'RIGHT_PRECEDENT' );
+        my $additional_mappings = $or_self->{searchengine}{elasticsearch}{additional_mappings} || {};
+        $mappings = $merge->merge($mappings, $additional_mappings);
+        print STDERR "create.mappings:\n".JSON::XS->new->convert_blessed->pretty->encode($mappings) if $debug;
+
+        # create
+        my $answer = $or_es->indices->create
+         (
+          index => $s_index,
+          body  => { mappings => $mappings },
+         );
+        print STDERR "create.answer:\n".Data::Dumper::Dumper($answer) if $debug;
+    }
+}
+
+sub sync_search_engine
+{
+    my ( $or_self, $b_force, $i_start, $i_count) = @_;
+
+    $i_start ||= 1;
+    $i_count ||= 10_000;
+
+    if ($or_self->{searchengine}{elasticsearch})
+    {
+        require BenchmarkAnything::Storage::Search::Elasticsearch;
+        my ($or_es, $s_index, $s_type) = BenchmarkAnything::Storage::Search::Elasticsearch::get_elasticsearch_client
+         (
+          {searchengine => $or_self->{searchengine}, ownjson => 1}
+         );
+        my $bulk = $or_es->bulk_helper(index => $s_index, type => $s_type);
+        my $i_count_datapoints = $or_self->{query}->select_count_datapoints->fetch->[0];
+
+        for (my $i = $i_start; $i <= $i_count_datapoints; $i += $i_count)
+        {
+            if ($b_force
+                # If the beginning or the end of a range do not exist we sync that window.
+                # Careful! This will not sync that window if there is a holein between! So
+                # if in doubt you better force a full sync.
+                or not $or_es->exists(index => $s_index, type => $s_type, id => $i)
+                or not $or_es->exists(index => $s_index, type => $s_type, id => $i+$i_count-1)
+                )
+            {
+                # Make sure to query the ::Backend::SQL store!
+                my $bmks = $or_self->get_full_benchmark_points($i, $i_count);
+                $bulk->index({ id => $_->{VALUE_ID}, source => $_}) foreach @$bmks;
+                $bulk->flush;
+            }
+        }
+    } else {
+            # Unsupported search engine
+    }
+}
+
 1;
 
 __END__
@@ -900,16 +1208,13 @@ way to the the database. A search function with complexe filters already exists.
 
 =head3 new
 
-=over 4
-
-=item
-
 Create a new B<BenchmarkAnything::Storage::Backend::SQL> object.
 
     my $or_bench = BenchmarkAnything::Storage::Backend::SQL->new({
         dbh    => $or_dbh,
         debug  => 0,
         config => YAML::Syck::LoadFile('~/conf/tapper_benchmark.conf'),
+        searchengine => ... # optional, see below at "Elasticsearch support"
     });
 
 =over 4
@@ -930,14 +1235,8 @@ written to STDOUT. The default is 0.
 
 =back
 
-=back
-
 =head3 add_single_benchmark
 
-=over 4
-
-=item
-
 Add one or more data points to a single benchmark to the database.
 
     my $b_success = $or_bench->add_single_benchmark({
@@ -983,8 +1282,6 @@ Ignore forgivable errors while writing.
 
 =back
 
-=back
-
 =head3 add_multi_benchmark
 
 Add one or more data points for multiple benchmarks to the database.
@@ -1248,11 +1545,20 @@ how many metrics, how many additional keys, are stored.
 
 =head3 get_single_benchmark_point
 
-Get a single data points from the database including all essential
-fields (NAME, VALUE, UNIT) and additional fields.
+Get a single data point from the database including all essential
+fields (NAME, VALUE, UNIT, VALUE_ID, CREATED) and all additional
+fields.
 
  my $point = $or_bench->get_single_benchmark_point($value_id);
 
+=head3 get_full_benchmark_points
+
+Get C<$count> data points from the database including all essential
+fields (NAME, VALUE, UNIT, VALUE_ID, CREATED) and all additional
+fields, beginning with C<$value_id>.
+
+ my $point = $or_bench->get_full_benchmark_points($value_id, $count);
+
 =head3 list_benchmark_names
 
 Get a list of all benchmark NAMEs, optionally matching a given pattern
@@ -1362,6 +1668,26 @@ isn't desired a false value must be passed.
 
 =back
 
+=head3 init_search_engine( $force )
+
+Initializes the configured search engine (Elasticsearch). If the index
+already exists it does nothing, except when you set C<$force> to a
+true value which deletes and re-creates the index. This is necessary
+for example to apply new type mappings.
+
+After a successful (re-)init you need to run C<sync_search_engine>.
+
+During (re-init) and sync you should disable querying by setting
+
+  searchengine.elasticsearch.enable_query: 0
+
+=head3 sync_search_engine( $force, $start, $count)
+
+Sync C<$count> (default 10000) entries from the relational backend
+into the search engine (Elasticsearch) for indexing, beginning at
+C<$start> (default 1). Already existing entries in Elasticsearch are
+skipped unless C<$force> is set to a true value.
+
 =head1 NAME
 
 BenchmarkAnything::Storage::Backend::SQL - Save and search benchmark points by database
@@ -1413,13 +1739,102 @@ Returns the list of operators supported by the query language. This is
 provided for frontend systems that support creating queries
 automatically.
 
+=head3 json_true
+
+Auxiliary function for Elasticsearch JSON data.
+
+=head3 json_false
+
+Auxiliary function for Elasticsearch JSON data.
+
+=head2 Elasticsearch support
+
+=head3 Config
+
+You can pass through a config entry for an external search engine
+(currently only Elasticsearch) to the constructor:
+
+    my $or_bench = BenchmarkAnything::Storage::Backend::SQL->new({
+        dbh    => $or_dbh,
+        searchengine => {
+          elasticsearch =>
+            #
+            # which index/type to use
+            index => "myapp",
+            type  => "benchmarkanything",
+            #
+            # queries use the searchengine
+            enable_query => 1,
+            #
+            # should each single added value be stored immediately
+            # (maybe there is another bulk sync mechanism)
+            index_single_added_values_immediately => 1,
+            #
+            # which nodes to use
+            nodes => [ 'localhost:9200' ],
+            #
+            # (OPTIONAL)
+            # Your additional application specific mappings, used when
+            # index is created.
+            #
+            # WARNING: You are allowed to overwrite the pre-defined
+            # defaults for the built-in fields (NAME, VALUE, VALUE_ID,
+            # UNIT, CREATED) as it is a RIGHT_PRECEDENT hash merge with
+            # your additional_mappings on the right side. So if you
+            # touch the internal fields you better know what you are
+            # doing.
+            additional_mappings => {
+                # type as defined above in elasticsearch.type
+                benchmarkanything => {
+                    # static key <properties>
+                    properties => {
+                        # field
+                        tapper_report => {
+                            type => long,
+                        },
+                        tapper_testrun => {
+                            type => long,
+                        },
+                        tapper_testplan => {
+                            type => long,
+                        },
+                    },
+                },
+            },
+        },
+    });
+
+With such a config and an already set up Elasticsearch you can use the
+lib as usual but it is handling all things transparently behind the
+scenes to index and query the data with Elasticsearch. The relational
+SQL storage is still used as the primary storage.
+
+=head3 Index
+
+When C<index_single_added_values_immediately> is set, every single
+added entry is fetched right after insert (to get all transformations
+and added metadata) and sent to elasticsearch for index.
+
+Please note, this immediate indexing adds an overhead to insert
+time. You could as well switch-off this setting and take care of
+indexing the data at another time. Then again, for instance the
+C<::Frontend::HTTP> already takes care of bulk-adding new data
+asynchronously, so the overhead should be hidden in there, so just
+switch-on the feature and don't worry too much.
+
+=head3 Search
+
+When C<enable_query> is set, the BenchmarkAnything queries are
+transformed into corresponding Elasticsearch queries, sent to
+Elastisearch, and the result is taken directly from its answers.
+
 =head1 AUTHOR
 
 Roberto Schaefer <schaefr@amazon.com>
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is Copyright (c) 2016 by Amazon.com, Inc. or its affiliates.
+This software is Copyright (c) 2017 by Amazon.com, Inc. or its affiliates.
 
 This is free software, licensed under:
 
